/* NET-DISPLAY.C - Program to print network parameters and other such info. */

/* Copyright (c) 1995, 1996 by Radford M. Neal 
 *
 * Permission is granted for anyone to copy, use, or modify this program 
 * for purposes of research or education, provided this copyright notice 
 * is retained, and note is made of any changes that have been made. 
 *
 * This program is distributed without any warranty, express or implied.
 * As this program was written for research purposes only, it has not been
 * tested to the degree that would be advisable in any important application.
 * All use of this program is entirely at the user's own risk.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "misc.h"
#include "log.h"
#include "prior.h"
#include "model.h"
#include "net.h"


/* MAIN PROGRAM. */

void main
( int argc,
  char **argv
)
{
  net_arch *a;
  model_specification *m;

  net_params  params, *w = &params;
  net_sigmas  sigmas, *s = &sigmas;

  log_file logf;
  log_gobbled logg;

  int sigmas_only;
  int params_only;
  int index;

  /* Look at arguments. */

  sigmas_only = 0;
  params_only = 0;

  if (argc>1 && (strcmp(argv[1],"-p")==0 || strcmp(argv[1],"-w")==0))
  { params_only = 1;
    argv += 1;
    argc -= 1;
  }
  else if (argc>1 && (strcmp(argv[1],"-h")==0 || strcmp(argv[1],"-s")==0))
  { sigmas_only = 1;
    argv += 1;
    argc -= 1;
  }

  index = -1;

  if (argc!=2 && argc!=3 
   || argc>2 && (index = atoi(argv[2]))<=0 && strcmp(argv[2],"0")!=0) 
  { fprintf(stderr,"Usage: net-display [ -p | -h ] log-file [ index ]\n");
    exit(1);
  }

  logf.file_name = argv[1];

  /* Open log file and read network architecture. */

  log_file_open (&logf, 0);

  log_gobble_init(&logg,0);
  net_record_sizes(&logg);

  while (!logf.at_end && logf.header.index<0)
  { log_gobble(&logf,&logg);
  }

  a = logg.data['A'];
  m = logg.data['M'];
  
  if (a==0)
  { fprintf(stderr,"No architecture specification in log file\n");
    exit(1);
  }

  s->total_sigmas = net_setup_sigma_count(a,m);
  w->total_params = net_setup_param_count(a);

  logg.req_size['S'] = s->total_sigmas * sizeof(net_sigma);
  logg.req_size['W'] = w->total_params * sizeof(net_param);

  /* Read the desired network from the log file. */

  if (index<0)
  { 
    log_gobble_last(&logf,&logg);

    if (logg.last_index<0)
    { fprintf(stderr,"No network in log file\n");
      exit(1);
    }

    index = logg.last_index;
  }
  else
  {
    while (!logf.at_end && logf.header.index!=index)
    { log_file_forward(&logf);
    }

    if (logf.at_end)
    { fprintf(stderr,"No network with that index is in the log file\n");
      exit(1);
    }

    log_gobble(&logf,&logg);
  }

  if (logg.index['W']!=index)
  { fprintf(stderr,"No weights stored for the network with that index\n");
    exit(1);
  }

  if (logg.index['S']!=index)
  { fprintf(stderr,"No sigmas stored for the network with that index\n");
    exit(1);
  }

  s->sigma_block = logg.data['S'];
  w->param_block = logg.data['W'];

  net_setup_sigma_pointers (s, a, m);
  net_setup_param_pointers (w, a);

  /* Print values of the parameters and hyperparameters, or whatever. */

  printf("\nNetwork in file \"%s\" with index %d%s\n", logf.file_name, index,
    sigmas_only ? " (sigmas only)" : params_only ? " (parameters only)" : "");

  if (params_only)
  { net_print_params(w,0,a,m);
  }
  else if (sigmas_only)
  { net_print_sigmas(s,a,m);
  }
  else
  { net_print_params(w,s,a,m);
  }

  printf("\n");

  exit(0);
}
